/* Front-end language hooks for the 99 language.
   Copyright (c) 2013 Free Software Foundation, Inc.
   Copyright (c) 2014 Christopher Smith

This file is part of g99.

g99 is free software: you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

g99 is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
g99. If not, see <http://www.gnu.org/licenses/>. */

#include "config.h"
#include "system.h"
#include "ansidecl.h"
#include "coretypes.h"
#include "tree.h"
#include "tree-iterator.h"
#include "dumpfile.h"
#include "gimple.h"
#include "cgraph.h"
#include "opts.h"
#include "options.h"
#include "flags.h"
#include "diagnostic-core.h"
#include "toplev.h"
#include "langhooks.h"
#include "langhooks-def.h"
#include "debug.h"

/* Required structures. */

struct GTY(()) lang_decl
{
  char dummy;
};

struct GTY(()) lang_type
{
  char dummy;
};

struct GTY(()) language_function
{
  char dummy;
};

struct GTY(()) lang_identifier
{
  struct tree_identifier common;
};

union GTY((desc ("0"),
           chain_next ("(union lang_tree_node *) TREE_CHAIN (&%h.generic)")))
      lang_tree_node
{
  union tree_node GTY((tag ("0"), desc ("tree_node_structure (&%h)"))) generic;
};

/* Global variables. */

static bool warn_foreign_comments = false;

static tree global_decls = NULL_TREE;

/* Language hook functions. */

unsigned int
ninetynine_option_lang_mask (void)
{
  return CL_ninetynine;
}

bool
ninetynine_handle_option (size_t code, const char *arg ATTRIBUTE_UNUSED,
                          int value, int kind ATTRIBUTE_UNUSED,
                          location_t loc ATTRIBUTE_UNUSED,
                          const struct cl_option_handlers *handlers
                              ATTRIBUTE_UNUSED)
{
  if (code == OPT_Wforeign_comments)
    {
      warn_foreign_comments = value != 0;
      return true;
    }
  else
    return false;
}

bool
ninetynine_init (void)
{
  build_common_tree_nodes (false, false);
  return true;
}

static void
check_file (const char *fname)
{
  FILE *in = fopen (fname, "r");
  if (in == NULL)
    fatal_error ("Can't open input file %s", fname);

  linemap_add (line_table, LC_ENTER, 0, fname, 1);
  unsigned int line = 0, column = 0;
  linemap_line_start (line_table, ++line, 0);

  /* Yes, this block is ugly; if anyone asks, tell them it's "optimised". */
  int c;
  while ((c = getc (in)) != EOF)
    {
      column++;
      switch (c)
        {
        case '/':
          c = getc (in);
          if (c == '/') /* C++/C99-style comment. */
            {
        case '\'': /* VB-style comment. */
              if (warn_foreign_comments)
                warning_at (linemap_position_for_column (line_table, column),
                            OPT_Wforeign_comments,
                            "Nonstandard comment format.");
        case '#': /* Perl/sh-style comment. */
              while ((c = getc (in)) != EOF && c != '\n')
                ;
              linemap_line_start (line_table, ++line, 0);
              column = 0;
            }
          else if (c == '*') /* C-style comment. */
            {
              source_location loc = linemap_position_for_column (line_table,
                                                                 column);
              if (warn_foreign_comments)
                warning_at (loc, OPT_Wforeign_comments,
                            "Nonstandard comment format.");
              column++;
              do
                {
                  while ((c = getc (in)) != '*')
                    {
                      column++;
                      if (c == '\n')
                        {
                          linemap_line_start (line_table, ++line, 0);
                          column = 0;
                        }
                      else if (c == EOF)
                        {
                          error_at (loc, "Syntax error.");
                          break;
                        }
                    }
                  column++;
                  while ((c = getc (in)) == '*')
                    column++;
                  if (c == '\n')
                    {
                      linemap_line_start (line_table, ++line, 0);
                      column = 0;
                    }
                }
              while (c != '/' && c != EOF);
              column++;
            }
          else
            {
              error_at (linemap_position_for_column (line_table, column),
                        "Syntax error.");
              while (c != EOF && c != '\n')
                c = getc (in);
              linemap_line_start (line_table, ++line, 0);
              column = 0;
            }
          break;
        case '\n':
          linemap_line_start (line_table, ++line, 0);
          column = 0;
          break;
        default:
          if (!ISSPACE (c))
            {
              error_at (linemap_position_for_column (line_table, column),
                        "Syntax error.");
              while (c != EOF && c != '\n')
                c = getc (in);
              linemap_line_start (line_table, ++line, 0);
              column = 0;
            }
        }
    }

  fclose (in);
}

static tree
string_cst (const char *str)
{
  return build_string_literal (strlen(str) + 1, str);
}

static tree
build_bottles_fn (void)
{
  tree params = tree_cons (NULL_TREE, integer_type_node,
                           tree_cons (NULL_TREE, void_type_node, NULL_TREE));
  tree fn_type = build_function_type (void_type_node, params);
  tree fn_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                             get_identifier ("bottles"), fn_type);

  DECL_CONTEXT (fn_decl) = NULL_TREE;
  TREE_STATIC (fn_decl) = true;
  TREE_PUBLIC (fn_decl) = true;

  /* Argument. */
  tree n_parm = build_decl (BUILTINS_LOCATION, PARM_DECL, get_identifier ("n"),
                            integer_type_node);
  DECL_ARG_TYPE (n_parm) = integer_type_node;
  DECL_CONTEXT (n_parm) = fn_decl;
  TREE_USED (n_parm) = true;
  DECL_ARGUMENTS (fn_decl) = n_parm;

  /* Return value. */
  tree ret = build_decl (BUILTINS_LOCATION, RESULT_DECL, NULL_TREE,
                         TREE_TYPE (fn_type));
  DECL_CONTEXT (ret) = fn_decl;
  DECL_ARTIFICIAL (ret) = true;
  DECL_IGNORED_P (ret) = true;

  DECL_RESULT (fn_decl) = ret;

  /* Declare the printf () function. */
  tree printf_fn_type = build_function_type (integer_type_node, NULL_TREE);
  tree printf_fn_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                                    get_identifier ("printf"), fn_type);
  DECL_EXTERNAL (printf_fn_type) = true;
  TREE_PUBLIC (printf_fn_type) = true;

  tree bl = build_block (NULL_TREE, NULL_TREE, fn_decl, NULL_TREE);
  DECL_INITIAL (fn_decl) = bl;
  TREE_USED (bl) = true;

  tree bind = build3 (BIND_EXPR, void_type_node, BLOCK_VARS (bl), NULL_TREE,
                      bl);
  TREE_SIDE_EFFECTS (bind) = true;

  tree stmts = alloc_stmt_list ();

  tree printf_0 = build_call_expr (printf_fn_decl, 1,
                                   string_cst ("no more bottles"));
  tree printf_1 = build_call_expr (printf_fn_decl, 1, string_cst ("1 bottle"));
  tree printf_n = build_call_expr (printf_fn_decl, 2,
                                   string_cst ("%d bottles"), n_parm);
  tree zero_test = build2 (EQ_EXPR, integer_type_node, n_parm,
                           build_int_cst (integer_type_node, 0));
  tree one_test = build2 (EQ_EXPR, integer_type_node, n_parm,
                          build_int_cst (integer_type_node, 1));
  tree conditional_expr = build3 (COND_EXPR, void_type_node, zero_test,
                                  printf_0,
                                  build3 (COND_EXPR, void_type_node,
                                          one_test, printf_1, printf_n));
  append_to_statement_list (conditional_expr, &stmts);

  BIND_EXPR_BODY (bind) = stmts;
  DECL_SAVED_TREE (fn_decl) = bind;

  return fn_decl;
}

static tree
build_main_fn (void)
{
  tree fn_type = build_function_type (integer_type_node, NULL_TREE);
  tree fn_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                             get_identifier ("main"), fn_type);

  DECL_CONTEXT (fn_decl) = NULL_TREE;
  TREE_STATIC (fn_decl) = true;
  TREE_PUBLIC (fn_decl) = true;
  DECL_ARGUMENTS (fn_decl) = NULL_TREE;

  /* Return value for main (). */
  tree main_ret = build_decl (BUILTINS_LOCATION, RESULT_DECL, NULL_TREE,
                              integer_type_node);
  DECL_CONTEXT (main_ret) = fn_decl;
  DECL_ARTIFICIAL (main_ret) = true;
  DECL_IGNORED_P (main_ret) = true;

  DECL_RESULT (fn_decl) = main_ret;

  /* Declare the bottles () function. */
  tree bottles_fn_type = build_function_type (void_type_node, NULL_TREE);
  tree bottles_fn_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                                     get_identifier ("bottles"), fn_type);
  DECL_EXTERNAL (bottles_fn_type) = true;
  TREE_PUBLIC (bottles_fn_type) = true;

  /* Declare the printf () function. */
  tree printf_fn_type = build_function_type (integer_type_node, NULL_TREE);
  tree printf_fn_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                                    get_identifier ("printf"), fn_type);
  DECL_EXTERNAL (printf_fn_type) = true;
  TREE_PUBLIC (printf_fn_type) = true;

  /* Variable declaration. */
  tree n_decl = build_decl (BUILTINS_LOCATION, VAR_DECL, NULL_TREE,
                            integer_type_node);

  tree bl = build_block (n_decl, NULL_TREE, fn_decl, NULL_TREE);
  DECL_INITIAL (fn_decl) = bl;
  TREE_USED (bl) = true;

  /* The bind expression contains the statements to execute. */
  tree bind = build3 (BIND_EXPR, void_type_node, BLOCK_VARS (bl), NULL_TREE,
                      bl);
  TREE_SIDE_EFFECTS (bind) = true;

  /* The statement list. */
  tree stmts = alloc_stmt_list ();

  /* n = 100; */
  tree init_n = build2 (MODIFY_EXPR, integer_type_node, n_decl,
                        build_int_cst (integer_type_node, 100));
  append_to_statement_list (init_n, &stmts);

  /* while (--n) { */
  tree while_stmts = alloc_stmt_list ();
  tree decr_n = build2 (PREDECREMENT_EXPR, integer_type_node, n_decl,
                        build_int_cst (integer_type_node, 1));
  tree test_expr = build2 (EQ_EXPR, integer_type_node, decr_n,
                           build_int_cst (integer_type_node, 0));
  tree break_expr = build1 (EXIT_EXPR, void_type_node, test_expr);
  append_to_statement_list (break_expr, &while_stmts);

  /* bottles (n); */
  tree bottles_n = build_call_expr (bottles_fn_decl, 1, n_decl);
  append_to_statement_list (bottles_n, &while_stmts);

  /* printf (" of beer on the wall, "); */
  tree printf_1 = build_call_expr (printf_fn_decl, 1,
                                   string_cst (" of beer on the wall, "));
  append_to_statement_list (printf_1, &while_stmts);

  /* bottles (n); */
  append_to_statement_list (bottles_n, &while_stmts);

  /* printf (" of beer.\nTake one down and pass it around, "); */
  tree printf_2 = build_call_expr (printf_fn_decl, 1,
      string_cst (" of beer.\nTake one down and pass it around, "));
  append_to_statement_list (printf_2, &while_stmts);

  /* bottles (n-1); */
  tree n_minus_1 = build2 (MINUS_EXPR, integer_type_node, n_decl,
                           build_int_cst (integer_type_node, 1));
  tree bottles_n_1 = build_call_expr (bottles_fn_decl, 1, n_minus_1);
  append_to_statement_list (bottles_n_1, &while_stmts);

  /* printf (" of beer.\n\n"); */
  tree printf_3 = build_call_expr (printf_fn_decl, 1,
                                   string_cst (" of beer on the wall.\n\n"));
  append_to_statement_list (printf_3, &while_stmts);

  /* } */
  tree while_expr = build1 (LOOP_EXPR, void_type_node, while_stmts);
  append_to_statement_list (while_expr, &stmts);

  /* printf("No more bottles of beer on the wall, "
   *        "no more bottles of beer.\n"
   *        "Go to the store and buy some more, "
   *        "99 bottles of beer on the wall.\n"); */
  tree printf_4 = build_call_expr (printf_fn_decl, 1,
      string_cst ("No more bottles of beer on the wall, "
                  "no more bottles of beer.\n"
                  "Go to the store and buy some more, "
                  "99 bottles of beer on the wall.\n"));
  append_to_statement_list (printf_4, &stmts);

  /* return 0; */
  tree set_ret = build2 (MODIFY_EXPR, TREE_TYPE (main_ret), main_ret,
                         build_int_cst (integer_type_node, 0));
  TREE_USED (set_ret) = true;
  tree ret_expr = build1 (RETURN_EXPR, void_type_node, set_ret);
  append_to_statement_list (ret_expr, &stmts);

  BIND_EXPR_BODY (bind) = stmts;
  DECL_SAVED_TREE (fn_decl) = bind;

  return fn_decl;
}

/* Language hook to read and parse the source file. */
void
ninetynine_parse_file (void)
{
  /* Read in the source file(s). */
  for (unsigned int i = 0; i < num_in_fnames; i++)
    check_file (in_fnames[i]);

  global_decls = chainon (build_main_fn (), global_decls);
  global_decls = chainon (build_bottles_fn (), global_decls);

  /* Dump out the translation unit. */
  FILE *tu_stream = dump_begin (TDI_tu, NULL);
  if (tu_stream)
    {
      dump_node (global_decls, 0, tu_stream);
      dump_end (TDI_tu, tu_stream);
    }

  /* Prepare declarations for middle-end. */
  for (tree t = global_decls; t; t = TREE_CHAIN (t))
    {
      gimplify_function_tree (t);
      cgraph_finalize_function (t, false);
    }
}

/* Unused; terminate with ICE. */
bool
ninetynine_global_bindings_p (void)
{
  gcc_unreachable ();
}

/* Unused; terminate with ICE. */
tree
ninetynine_pushdecl (tree decl ATTRIBUTE_UNUSED)
{
  gcc_unreachable ();
}

/* Return the global declarations. */
tree
ninetynine_getdecls (void)
{
  return global_decls;
}

/* Unused; terminate with ICE. */
tree
ninetynine_type_for_mode (enum machine_mode mode ATTRIBUTE_UNUSED,
                          int unsignedp ATTRIBUTE_UNUSED)
{
  gcc_unreachable ();
}

/* Unused; terminate with ICE. */
tree
ninetynine_type_for_size (unsigned precision ATTRIBUTE_UNUSED,
                          int unsignedp ATTRIBUTE_UNUSED)
{
  gcc_unreachable ();
}

/* Language hooks time! */

#undef LANG_HOOKS_NAME
#undef LANG_HOOKS_OPTION_LANG_MASK
#undef LANG_HOOKS_HANDLE_OPTION
#undef LANG_HOOKS_INIT
#undef LANG_HOOKS_PARSE_FILE
#undef LANG_HOOKS_GLOBAL_BINDINGS_P
#undef LANG_HOOKS_PUSHDECL
#undef LANG_HOOKS_GETDECLS
#undef LANG_HOOKS_TYPE_FOR_MODE
#undef LANG_HOOKS_TYPE_FOR_SIZE

#define LANG_HOOKS_NAME              "GNU 99"
#define LANG_HOOKS_OPTION_LANG_MASK  ninetynine_option_lang_mask
#define LANG_HOOKS_HANDLE_OPTION     ninetynine_handle_option
#define LANG_HOOKS_INIT              ninetynine_init
#define LANG_HOOKS_PARSE_FILE        ninetynine_parse_file
#define LANG_HOOKS_GLOBAL_BINDINGS_P ninetynine_global_bindings_p
#define LANG_HOOKS_PUSHDECL          ninetynine_pushdecl
#define LANG_HOOKS_GETDECLS          ninetynine_getdecls
#define LANG_HOOKS_TYPE_FOR_MODE     ninetynine_type_for_mode
#define LANG_HOOKS_TYPE_FOR_SIZE     ninetynine_type_for_size

struct lang_hooks lang_hooks = LANG_HOOKS_INITIALIZER;

/* GGC stuff. */

#include "gt-ninetynine-991.h"
#include "gtype-ninetynine.h"
